"""
data.py - the pinda database
"""
import os
import yaml
import glob
import shutil
import subprocess
import requests
import pinda


def reload_database():
    global database
    with open(pinda.PINDA_DEFAULT_DATABASE) as f:
        database = yaml.load(f, Loader=yaml.SafeLoader)
    for databasefile in glob.glob(pinda.PINDA_CONFIGDIR + '/*.y*ml'):
        if databasefile != pinda.PINDA_DEFAULT_DATABASE:
            with open(databasefile,'r') as f:
                d = yaml.load(f, Loader=yaml.SafeLoader)
                for k in d:
                    database[k] = d[k]

def update():
    """
    Download a fresh copy of _default.yaml from the pinda repository
    """
    result = requests.get('http://claughton.bitbucket.io/resources/pinda/_default.yaml')
    with open(pinda.PINDA_DEFAULT_DATABASE, 'w') as f:
        f.write(result.text)

def update_from(yaml_file, overwrite=False):
    """
    Update or add an entry to the database.
    """
    with open(yaml_file) as f:
        try:
            new_packages = yaml.load(f, Loader=yaml.SafeLoader)
        except:
            raise TypeError('Error: cannot parse {} - check format'.format(yaml_file))
    for package in new_packages:
        for k in ['name', 'version', 'package', 'repository', 
              'description', 'info']:
            if not k in new_packages[package]:
                raise KeyError('Error in description of package {}: key {} is required'.format(package, k))

    global database
    reload_database()
    for package in new_packages:
        if package in database and not overwrite:
            raise ValueError('Error - {name} {version} is already in the database'.format(**database[package]))
    shutil.copy(yaml_file, pinda.PINDA_CONFIGDIR)
template = """
# A template for a file to add a new application to your local database of
# pinda-installable applications. This file is written in YAML format.
# Once complete, use 'pinda update' to update the database.
#
id: <a string made from <name>_<version> values below, e.g. gromacs_2018>
  name: <name of the application>
  version: <version of the application>
  package: <name of the underlying python package (in setup.py)>
  repository: <repository link, e.g. git+https://bitbucket.org/<user>/<repo>.git@<branch>
  description: <one-line summary of the application>
  info: |
    <Multi-line information about the application 
    and distribution.>
"""

default_database = {
            'ambertools_18':
            {
             'name': 'ambertools',
             'version': '18',
             'package': 'amberdocker',
             'repository': 'git+https://bitbucket.org/claughton/amber_docker.git@v18-tools',
             'description': 'AmberTools version 18',
             'info': '''
AmberTools version 18

This pinda distribution provides the following commands:

    "sander"      : the Amber/Ambertools "sander" command
    "amber-shell" : drops you into a bash shell from which all other
                    AmberTools commands are accessible (e.g. tleap).

For full license and usage information, please visit http://ambermd.org

'''
            },
            'fpocket_3.0':
            {
             'name': 'fpocket',
             'version': '3.0',
             'package': 'fpocketdocker',
             'repository': 'git+https://bitbucket.org/claughton/fpocket_docker.git',
             'description': 'FPocket version 3.0',
             'info': '''
FPocket version 3.0

This pinda distribution provides the following commands:

    fpocket:  the original pocket prediction on a single protein structure 
    mdpocket: extension of fpocket to analyse conformational ensembles of 
              proteins (MD trajectories for instance) 
    dpocket: extract pocket descriptors 
    tpocket: test your pocket scoring function

For full license and usage instructions, please visit https://github.com/Discngine/fpocket

'''
            },
            'gromacs_2019':
            {
             'name': 'gromacs',
             'version': '2019',
             'package': 'gromacsdocker',
             'repository': 'git+https://bitbucket.org/claughton/gromacs_docker.git@v2019',
             'description': 'Gromacs version 2019',
             'info': '''
Gromacs version 2019

This pinda distribution provides the following Gromacs commands:

    "gmx"        : The Gromacs "gmx" command
    "gmx-select" : A command to set the instruction set for optimal 
                   performance

For full license and usage instructions, please visit http://gromacs.org

'''
            },
            'gromacs_2019-cuda':
            {
             'name': 'gromacs',
             'package': 'gromacsdocker',
             'version': '2019-cuda',
             'repository': 'git+https://bitbucket.org/claughton/gromacs_docker.git@v2019-cuda',
             'description': 'Gromacs version 2019 with CUDA support',
             'info': '''
Gromacs version 2019 with CUDA support

This pinda distribution provides the following Gromacs commands:

    "gmx"        : The Gromacs "gmx" command
    "gmx-select" : A command to set the instruction set for optimal 
                   performance

For full license and usage instructions, please visit http://gromacs.org

'''
            },
            'obabel_2.4.1':
            {
             'name': 'obabel',
             'version': '2.4.1',
             'package': 'babeldocker',
             'repository': 'git+https://bitbucket.org/claughton/babel_docker.git',
             'description': 'Open Babel version 2.4.1',
             'info': '''
Open Babel version 2.4.1

This pinda distribution provides the following OpenBabel commands:

    "obabel"        : The OpenBabel "obabel" command

For full license and usage instructions, please visit http://openbabel.orgg

'''
            },
            'vina_1.1.2':
            {
             'name': 'vina',
             'version': '1.1.2',
             'package': 'vinadocker',
             'repository': 'git+https://bitbucket.org/claughton/vina_docker.git',
             'description': 'AutoDock Vina and selected AutoDock Tools',
             'info': '''
AutoDock Vina and selected AutoDock Tools

This pinda distribution provides the following commands from AutoDock Vina
and AutoDock tools:

    "vina" : The "vina" command from AutoDock Vina 1.1.2
    "adt"  : An interface to selected Python tools from AutoDock Tools
             version 4.2.6, including those for preparing PDBQT format
             input files for AutoDock Vina

For full license and usage instructions, please visit http://vina.scripps.edu
and http://autodock.scripps.edu

'''
            },
            'procheck_3.5.4':
            {
             'name': 'procheck',
             'version': '3.5.4',
             'package': 'procheckdocker',
             'repository': 'git+https://bitbucket.org/claughton/procheck_docker.git@v3.5.4',
             'description': 'PROCHECK protein structure quality assessment tools',
             'info': '''
PROCHECK protein structure quaklity assessment tools v 3.5.4


This pinda distribution provides the following PROCHECK commands:

    "procheck" : The PROCHECK "procheck" command.
    "gfac2pdb" : The PROCHECK "gfac2pdb" utility.

Note that the NMR-related tools in the full PROCHECK package are not included.


Users must complete and return the confidentiality agreement at https://www.ebi.ac.uk/thornton-srv/software/PROCHECK/ before use. 
'''
            },
            'propka_3.1.0':
            {
             'name': 'propka',
             'version': '3.1.0',
             'package': 'PROPKA',
             'repository': 'git+https://github.com/jensengroup/propka-3.1.git',
             'description': 'PROPKA version 3.1',
             'info': '''
PROPKA protein residue pKa predictor version 3.1


This pinda distribution provides the following PROPKA command:

    "propka31" : The PROPKA "propka31" command.

See https://github.com/jensengroup/propka-3.1 for full usage instructions, and appropriate citations.
'''
            },
            'reduce_3.3.160602':
            {
             'name': 'reduce',
             'version': '3.3.160602',
             'package': 'REDUCE',
             'repository': 'git+https://bitbucket.org/claughton/reduce_docker.git@v3.3.160602',
             'description': 'REDUCE version 3.3.160602',
             'info': '''
REDUCE utility to add hydrogen atoms to protein PDB files version 3.3.160602


This pinda distribution provides the following REDUCE command:

    "reduce" : The REDUCE "reduce" command.

See http://kinemage.biochem.duke.edu/software/reduce.php for full usage instructions, and appropriate citations.
'''
            },
            'NAMD_2.12':
            {
             'name': 'NAMD',
             'version': '2.12',
             'package': 'namddocker',
             'repository': 'git+https://bitbucket.org/claughton/namd_docker.git@v2.12',
             'description': 'NAMD parallel molecular dynamics code, version 2.12',
             'info': '''

    NAMD parallel molecular dynamics code, version 2.12


    This pinda distribution provides the following NAMD commands:

        "namd2" : The NAMD "namd2" command.

    Users must agree to the license conditions at https://www.ks.uiuc.edu/Research/namd/license.html before use. 
    See https://www.ks.uiuc.edu/Research/namd for full usage instructions.

'''
            },
           }

if not os.path.exists(pinda.PINDA_CONFIGDIR):
    os.mkdir(pinda.PINDA_CONFIGDIR)

if not os.path.exists(pinda.PINDA_DEFAULT_DATABASE):
    try:
        update()
    except:
        with open(pinda.PINDA_DEFAULT_DATABASE, 'w') as f:
            yaml.dump(default_database, f, default_flow_style=False)
reload_database()




"""
commands.py: pinda commands
"""
import subprocess
from . import utilities
from . import data

def list():
    """
    List available packages
    """
    available = utilities.available_packages()
    installed = utilities.installed_packages()
    result = ''
    for package in available:
        if package in installed:
            line = '* '
        else:
            line = '  '
        line += '{name:10s} {version:10s} {description}\n'.format(**package)
        result += line
    return result

def install(name, version, sudo=False, user=True):
    """
    Install a package
    """
    if utilities.is_installed(name, version):
        return
    r = utilities.available_packages(name, version)
    if len(r) == 0:
        raise ValueError('Error: {} {} is not available'.format(name, version))
    r = r[0]
    pip = utilities.pip_command()
    if pip is None:
        raise RuntimeError('Error - cannot find pip or pip3 command')
    r['pip_command'] = pip
    if sudo:
        command = 'sudo {pip_command} install {repository}'.format(**r)
    else:
        if user:
            command = '{pip_command} install {repository} --user'.format(**r)
        else:
            command = '{pip_command} install {repository}'.format(**r)
    result = subprocess.check_output(command, shell=True, 
                                     universal_newlines=True, 
                                     stderr = subprocess.STDOUT)
    
def uninstall(name, version, sudo=False):
    """
    Uninstall a package
    """
    if not utilities.is_installed(name, version):
        return True
    r = utilities.available_packages(name, version)[0]
    pip = utilities.pip_command()
    if pip is None:
        raise RuntimeError('Error - cannot find pip or pip3 command')
    r['pip_command'] = pip
    if sudo:
        command = 'sudo {pip_command} uninstall -y{package}'.format(**r)
    else:
        command = '{pip_command} uninstall -y {package}'.format(**r)
    result = subprocess.check_output(command, shell=True, 
                                     universal_newlines=True, 
                                     stderr = subprocess.STDOUT)

def template():
    """
    Returns a template for a YAML file to define a new database entry.
    """
    return data.template

def update_from(yaml_file, overwrite=False):
    """
    Update the local database of available packages.

    The atgument is the name of a YAML format file that contains the 
    specification for the new database entry. A template for this can
    be generated using the "pinda template" command.
    """
    data.update_from(yaml_file, overwrite=overwrite)

def update():
    """
    Update the pinda default database.

    Downloads a fresh copy of '_default.yaml' into $HOME/.pinda.
    """
    data.update()

def info(name, version):
    """
    Returns full information about a package
    """
    r = utilities.available_packages(name, version)
    if len(r) == 0:
        raise ValueError('Error: {} {} is not available'.format(name, version))
    r = r[0]
    return r['info']



import pytest
from pinda import utilities, commands

def test_docker_available():
    result = utilities.docker_installed()
    assert result is True

def test_available_all():
    result = utilities.available_packages()
    assert type(result) == list

def test_available_specific():
    result = utilities.available_packages('gromacs')
    assert len(result) == 2
    for r in result:
        assert r['name'] == 'gromacs'

def test_available_specific_version():
    result = utilities.available_packages('gromacs', '2019')
    assert len(result) == 1
    for r in result:
        assert r['version'] == '2019'
    
def test_installed_packages():
    commands.install('gromacs', '2019')
    result = utilities.installed_packages('gromacs', '2019')
    assert len(result) == 1
    result = utilities.is_installed('gromacs', '2019')
    assert result == True
    result = utilities.is_installed('gromacs', '2019-cuda')
    assert result == False



"""
utilities.py: pinda utility functions
"""
import subprocess
import os

from .data import database
def available_packages(name=None, version=None):
    result = []
    if name is None:
        target_id = '_'
    elif version is None:
        target_id = name + '_'
    else:
        target_id = name + '_' + str(version)
    ids = [i for i in database]
    ids.sort()
    for id in ids:
        if name is None and version is None:
            result.append(database[id])
        elif version is None:
            if target_id == id[:len(target_id)]:
                result.append(database[id])
        else:
            if target_id == id:
                result.append(database[id])
    return result

def is_available(name, version):
    return len(available_packages(name, version)) == 1

def installed_packages(name=None, version=None):
    available = available_packages(name, version)
    if len(available) == 0:
        return []
    installed = []
    DEVNULL = open(os.devnull, 'w')
    pip = pip_command()
    if pip is None:
        raise RuntimeError('Error - cannot find pip or pip3 command')
    listing = subprocess.check_output('{} list'.format(pip), universal_newlines=True, shell=True, stderr=DEVNULL).split('\n')
    for line in listing:
        words = line.split()
        if len(words) == 2:
            package = words[0]
            version = words[1]
            if version[0] == '(' and version[-1] == ')':
                version = version[1:-1]
            for p in available:
                if p['package'] == package and p['version'] == version:
                    installed.append(p)
    return installed

def is_installed(name, version):
    return len(installed_packages(name, version)) == 1

def docker_installed():
    try:
        result = subprocess.check_output('which docker', shell=True, 
                                          universal_newlines=True, 
                                          stderr=subprocess.STDOUT)
        return True
    except subprocess.CalledProcessError:
        return False

def pip_command():
    try:
        result = subprocess.check_output('which pip', shell=True,
                                         universal_newlines=True,
                                         stderr=subprocess.STDOUT)
        return 'pip'
    except subprocess.CalledProcessError:
        pass
    try:
        result = subprocess.check_output('which pip3', shell=True,
                                         universal_newlines=True,
                                         stderr=subprocess.STDOUT)
        return 'pip3'
    except subprocess.CalledProcessError:
        return None



import pytest
from pinda import commands, data

def test_install():
    commands.install('fpocket', '3.0')
    with pytest.raises(ValueError):
        commands.install('pocket', '3.0')
        commands.install('fpocket', '2.0')

def test_uninstall():
    commands.uninstall('fpocket', '3.0')
    commands.uninstall('pocket', '3.0')
    commands.uninstall('fpocket', '2.0')

def test_update():
    data.update_from('test/data/good_template.yaml')
    with pytest.raises(ValueError):
        data.update_from('test/data/good_template.yaml')
    data.update_from('test/data/good_template.yaml', overwrite=True)
